@@ -3,9 +3,6 @@ The contents of this file are focused on the Timestep class, which is used for s
 imported data from N-body output files.
 '''
 
-#TODO: UNIT TESTS
-#TODO: Create sort function that sorts based on id (not really importamt)
-
 #===============================================================================
 # IMPORTS
 #===============================================================================
@@ -47,7 +44,7 @@ class AttrDict(dict):
 
 class Timestep():
 
-    def __init__(self, typ=list(), id_val=list(), x=list(), y=list(), z=list(), vx=list(), vy=list(), vz=list(), mass=list(), center_of_mass=[0, 0, 0], center_of_momentum=[0, 0, 0], potential=None, time=None, nbody=None):
+    def __init__(self, typ=list(), id_val=list(), x=list(), y=list(), z=list(), vx=list(), vy=list(), vz=list(), mass=list(), center_of_mass=[0, 0, 0], center_of_momentum=[0, 0, 0], potential=None, time=None, nbody=None, **kwargs):
         #all this is typically specified by the read_output function in output_handler
         #read_output is the preferred way to input data to this data structure
         #but if you're really feeling adventurous you can always do it yourself
@@ -86,6 +83,12 @@ class Timestep():
         self.center_of_mass = center_of_mass
         self.center_of_momentum = center_of_momentum
 
+        # Store any additional quantities provided via kwargs
+        self._provided_vals = []
+        for kw in kwargs.keys():
+            self[kw] = np.array(kwargs[kw]) 
+            self._provided_vals.append(kw)
+
         #-----------------------------------------------------------------------
         # HOUSEKEEPING
         #-----------------------------------------------------------------------
@@ -283,13 +286,16 @@ class Timestep():
         self.msol = self.mass * struct_to_sol
 
         #position information
-        self.r = (self.x**2 + self.y**2 + self.z**2)**0.5
+        if not 'r' in self._provided_vals:
+            self.r = (self.x**2 + self.y**2 + self.z**2)**0.5
         self.dist = ((self.x + 8)**2 + self.y**2 + self.z**2)**0.5
         self.R = (self.x**2 + self.y**2)**0.5
 
         #galactic coordinate information
-        self.l = np.arctan2(self.y, self.x + 8)*180/np.pi
-        self.b = np.arcsin(self.z/self.dist)*180/np.pi
+        if not 'l' in self._provided_vals:
+            self.l = np.arctan2(self.y, self.x + 8)*180/np.pi
+        if not 'b' in self._provided_vals:
+            self.b = np.arcsin(self.z/self.dist)*180/np.pi
 
         #ICRS information
         c = SkyCoord(l=self.l*u.degree, b=self.b*u.degree, frame='galactic')
@@ -310,8 +316,9 @@ class Timestep():
         self.ltot = (self.lx**2 + self.ly**2 + self.lz**2)**0.5
 
         #velocity information
-        self.vgsr = ((self.x+8)*self.vx + self.y*self.vy + self.x*self.vz)/self.dist
-        self.vlos = self.vgsr - 10.1*np.cos(self.b*np.pi/180)*np.cos(self.l*np.pi/180) - 224*np.cos(self.b*np.pi/180)*np.sin(self.l*np.pi/180) - 6.7*np.sin(self.b*np.pi/180)
+        self.vgsr = ((self.x+8)*self.vx + self.y*self.vy + self.z*self.vz)/self.dist
+        if not 'vlos' in self._provided_vals:
+            self.vlos = self.vgsr - 10.1*np.cos(self.b*np.pi/180)*np.cos(self.l*np.pi/180) - 224*np.cos(self.b*np.pi/180)*np.sin(self.l*np.pi/180) - 6.7*np.sin(self.b*np.pi/180)
         self.vrad = (self.x*self.vx + self.y*self.vy + self.z*self.vz)/self.r
         self.vrot = self.lz/(self.x**2 + self.y**2)**0.5 #should have a factor of cos(theta) I think? Need to define what I mean exactly by this quantity
                                                          #should also make a polar velocity if I'm going to do that
@@ -331,7 +338,8 @@ class Timestep():
         if verbose:
             print('Calculating proper motion values...')
 
-        self.pmra, self.pmdec = get_rvpm(self.ra, self.dec, self.dist, self.vx, self.vy, self.vz)[1:]
+        if not 'pmra' in self._provided_vals or not 'pmdec' in self._provided_vals:
+            self.pmra, self.pmdec = get_rvpm(self.ra, self.dec, self.dist, self.vx, self.vy, self.vz)[1:]
         #already get rv from vlos, don't need to save it as something else
         self.pmtot = (self.pmra**2 + self.pmdec**2)**0.5
         #4.848e-6 is arcsec->rad, 3.086e16 is kpc->km, and 3.156e7 is sidereal yr -> seconds
@@ -394,6 +402,14 @@ class Timestep():
         if auto_update:
             self.update()
 
+    #cut out the rows with the given indices
+    def remove(self, indices):
+        #indices: the indices you want to take, must be array-like
+        for key in self:
+            self[key] = np.delete(self[key], indices)
+        if auto_update:
+            self.update()
+
     #cut the data to only include particles with id in ids
     def take_by_id(self, ids):
         #ids: the ids of the particles you want to take, must be array-like
@@ -417,35 +433,29 @@ class Timestep():
         Timestep1 = self.copy()
         Timestep2 = self.copy()
 
+        #these automatically update both timesteps
         Timestep1.cut_last_n(len(self) - n)
         Timestep2.cut_first_n(n)
 
-        if auto_update:
-            Timestep1.update()
-            Timestep2.update()
-
         return Timestep1, Timestep2
 
     #splits the Timestep into a list of new Timestep structures,
     #where the Timestep is split every time ID wraps back to zero
-    #TODO: the dwarf id start at 1, not 0
-    def split_at_id_wrap(self):
+    def split_at_id_wrap(self, val=1):
+        #val [int]: the index that is looked for when splitting the timesteps
+        #           (this will almost always be 1 but the toggle is here now just in case)
 
         outlist = []
-        indices = np.where(self.id==0)[0] #1D list of arrays
+        indices = np.where(self.id==val)[0] #1D list of arrays
 
         Timestep2 = self.copy()
         i = 1
         while i < len(indices): #pseudo-recursive
-            Timestep1, Timestep2 = Timestep2.split(indices[i] - indices[i-1])
+            Timestep1, Timestep2 = Timestep2.split(indices[i] - indices[i-1]) #this automatically updates each timestep
             outlist.append(Timestep1)
             i += 1
         outlist.append(Timestep2)
 
-        if auto_update:
-            for t in outlist:
-                t.update()
-
         return outlist
 
     #append a Timestep object onto this one
@@ -459,7 +469,7 @@ class Timestep():
 
     #append the nth item of another Timestep object onto this one
     #WARNING: Extremely time intensive if used repeatedly, np.append is not a speedy function
-    #   you should always try to append a timestep instead if possible
+    #   you should always try to append_timestep() instead if possible
     def append_point(self, t, n=0, id=None):
         #t: the Timestep object with the item being appended
         #n (optional): the index of the item in t to be appended
@@ -512,38 +522,30 @@ class Timestep():
         for i in data_list:
             self.append_timestep(i)
 
-        if auto_update:
-            self.update()
-
     #cut the Timestep so only baryons (typ==0) remain
     def only_baryons(self):
-        #find indices of particles with typ==0
-        ind = []
-        for i, t in zip(range(len(self.typ)), self.typ): #ugly but fast I think
-            if t==0:
-                ind.append(i)
-
-        #cut the data
-        self.take(ind)
+        #nasty one liner
+        self.take(np.nonzero((self.typ == 0).astype(int))[0])
 
     #cut the Timestep so only dark matter particles (typ==1) remain
     def only_dark_matter(self):
-        #find indices of particles with typ==1
-        ind = []
-        for i, t in zip(range(len(self.typ)), self.typ): #ugly but fast I think
-            if t==0:
-                ind.append(i)
-
-        #cut the data
-        self.take(ind)
+        #nasty one liner
+        self.take(np.nonzero((self.typ == 1).astype(int))[0])
 
     #---------------------------------------------------------------------------
     # OUTPUT
     #---------------------------------------------------------------------------
 
-    def print_particle(self, n, dec=8):
+    #"fancy" output for showing properties of a given particle
+    def print_particle(self, n, dec=8, use_id=False):
         #n (int): the location of the particle that you want the information for
-        #TODO: allow for using the id to find the particle
+
+        if use_id:
+            tmp = np.where(self.id == n)[0]
+            if len(tmp) == 0:
+                raise Exception('Particle with given ID was not found.')
+            n = tmp[0]
+
         print('Printing data for Particle '+str(n)+':')
 
         outstr = '('
@@ -579,8 +581,7 @@ class Timestep():
     #---------------------------------------------------------------------------
 
     #make an n-dimensional rectangular cut on the data
-    #TODO: make an inverted method, i.e. cut out things within the bounds
-    def subset_rect(self, axes, bounds):
+    def subset_rect(self, axes, bounds, inverse=False):
         #axes ([str, ...]): the parameters/values that you are cutting on. Input as a list of strings
         #   The strings must be in self.index_list
         #bounds ([(float, float), ...]): The boundary conditions that you are cutting on
@@ -588,9 +589,6 @@ class Timestep():
         #   If you wish to only declare a minimum or maximum, then leave the other value as None
         #
         #Both axes and bounds can be given as a tuple instead of a list
-        #
-        #This is substantially faster than the previous subset function, which performed like 30
-        #np.appends for every index that fit the criteria in each axis.
 
         if type(axes) not in [type([]), type((1,1))]: #make the input compatible if it is only 1 axis
             raise Exception('Axes must be a list or a tuple, but was of type ' + str(type(axes)))
@@ -612,18 +610,16 @@ class Timestep():
             indices = np.intersect1d(indices, np.intersect1d(np.where(self[a] > b[0]), np.where(self[a] < b[1])))
 
         #cut the sample down to the indices that work
-        self.take(indices)
-        self.update()
+        if inverse:
+            self.remove(indices)
+        else:
+            self.take(indices)
 
-    #TODO: make n-dimensional like subset_rect
-    #make a circular cut of the data in 2 dimensions
-    def subset_circ(self, axes, rads, centers):
+    #make an elliptical cut of the data in n dimensions
+    def subset_circ(self, axes, rads, centers, inverse=False):
         #axs ([str]): the axes to cut on
         #rads ([float]): The radiii of the circular cut (along each axis)
         #center ([float], len=#axes): The center around which to make the circular cut
-        #
-        #This is substantially faster than the previous subset function, which performed like 30
-        #np.appends for every index that fit the criteria in each axis.
 
         #get the indices that lie within the cut
         dist = 0
@@ -633,8 +629,10 @@ class Timestep():
         indices = np.where(dist < 1)[0]
 
         #cut the sample down to the indices that work
-        self.take(indices)
-        self.update()
+        if inverse:
+            self.remove(indices)
+        else:
+            self.take(indices)
 
     #cut the Timestep instance to n random stars from the Timestep instance
     #uses the reservoir algorithm for single-pass random sampling
@@ -716,7 +714,7 @@ def get_self_energies(t):
 
     #compute the kinetic energy of each particle (within reference frame of
     #the COM's of the particles)
-    kin_energies = t.vx**2 + t.vy**2 + t.vz**2
+    kin_energies = 0.5*(t.vx**2 + t.vy**2 + t.vz**2)
 
     energies = pot_energies + kin_energies
 
@@ -728,7 +726,7 @@ def get_self_energies(t):
 #written by Eric Mendelsohn, 2021
 #adapted for mwahpy by Tom Donlon, 2021
 
-#TODO: Can have trouble finding progenitors around poles. If progenitor is
+#NOTE: Can have trouble finding progenitors around poles. If progenitor is
 #   above 60 deg or so, could apply arbitrary spherical rotation to bring it back
 #   into a spot that the algorithm handles better
 def find_progenitor(t, ran=100.):
@@ -794,3 +792,52 @@ def find_progenitor(t, ran=100.):
     pos = [x_val,y_val,z_val]
 
     return pos
+
+#computes the amount of bound mass (split up by DM and baryons) around a dwarf
+#assumes that all of the particles in the timestep belong to the dwarf
+#adapted from code written by Hiroka Warren
+
+#output = total bound mass, baryon bound mass, DM bound mass (in structure masses)
+
+def get_bound_mass(t):
+    #t: the timestep
+
+    self_energies = get_self_energies(t)
+
+    bound = (self_energies < 0)
+    bary = (t.typ == 0)
+    dm = (t.typ == 1)
+
+    return np.sum(t.mass[bound]), np.sum(t.mass[bound*bary]), np.sum(t.mass[bound*dm])
+
+
+#checks the level of incest in each particle and provides a dict of all
+#particles with incest at least the specified level, and their counterparts
+#(incest is when two particles share at least one Cartesian position. I didn't
+# come up with the name)
+
+#the dict has the format {id: [list, of, all, incestuous, ids]}
+
+#heads up this runs as O(n) so if you have a very large timestep, be ready to wait around for a while
+
+def check_for_incest(t, lvl=1):
+    #lvl: how many cartesian coordinates need to be shared for the program to include a particle as incestuous
+
+    out_dict = {}
+
+    for i in range(len(t)):
+        dx = t.x - t.x[i]
+        dy = t.y - t.y[i]
+        dz = t.z - t.z[i]
+
+        tot_lvl = (dx == 0).astype(int) + (dy == 0).astype(int) + (dz == 0).astype(int)
+        inds = np.where(tot_lvl > lvl)[0]
+
+        if len(inds) > 0:
+            out_dict[i] = inds 
+
+    if verbose:
+        print(f'{len(out_dict)} particles were incestuous in at least {lvl} coordinates')
+
+    return out_dict
+
